use std::{cmp::Ordering, collections::BinaryHeap, ops::Range};

use crate::grid::CellType;

pub trait PathfinderTarget {
	fn get_edges(&self, node_id: usize) -> Vec<usize>;
	fn node_range(&self) -> Range<usize>;
	fn hscore(&self, node_id: usize) -> usize;
	fn distance(&self, node_a: usize, node_b: usize) -> usize;
}

pub struct AStar {}

pub trait Pathfinder {
	fn pathfind<T>(from: usize, to: usize, target: &T) -> (Vec<(usize, usize)>, Vec<usize>)
	where
		T: PathfinderTarget;
}

pub trait PathfinderCell {
	fn walkable(&self) -> bool;
	fn celltype(&self) -> CellType;
}

#[derive(PartialEq, Eq)]
struct Edge {
	fscore: usize,
	distance: usize,
	from: usize,
	to: usize,
}
impl PartialOrd for Edge {
	fn partial_cmp(&self, other: &Self) -> Option<Ordering> {
		Some(match self.fscore.cmp(&other.fscore) {
			Ordering::Less => Ordering::Greater,
			Ordering::Equal => Ordering::Equal,
			Ordering::Greater => Ordering::Less,
		})
	}
}
impl Ord for Edge {
	fn cmp(&self, other: &Self) -> Ordering {
		match self.fscore.cmp(&other.fscore) {
			Ordering::Less => Ordering::Greater,
			Ordering::Equal => Ordering::Equal,
			Ordering::Greater => Ordering::Less,
		}
	}
}
#[derive(Clone)]
struct Node {
	distance: usize,
	visited: bool,
	previous: Option<usize>,
	id: usize,
}
impl Pathfinder for AStar {
	fn pathfind<T>(from: usize, to: usize, target: &T) -> (Vec<(usize, usize)>, Vec<usize>)
	where
		T: PathfinderTarget,
	{
		let range = target.node_range();
		let mut nodes = vec![
			Node {
				distance: usize::max_value(),
				visited: false,
				previous: None,
				id: 0,
			};
			range.len()
		];
		nodes[from - range.start] = Node {
			distance: 0,
			visited: true,
			previous: None,
			id: from,
		};
		let mut edges = BinaryHeap::<Edge>::new();
		for edge in target.get_edges(from) {
			edges.push(Edge {
				fscore: target.hscore(edge),
				distance: target.distance(from, edge),
				from,
				to: edge,
			});
		}
		while let Some(current) = edges.pop() {
			if nodes[current.to - range.start].visited {
				continue;
			}
			nodes[current.to - range.start] = Node {
				distance: current.distance,
				visited: true,
				previous: Some(current.from),
				id: current.to,
			};
			if current.to == to {
				break;
			}
			for neighbor in target.get_edges(current.to) {
				edges.push(Edge {
					fscore: current.distance
						+ target.hscore(neighbor)
						+ target.distance(current.to, neighbor),
					distance: current.distance + target.distance(current.to, neighbor),
					from: current.to,
					to: neighbor,
				});
			}
		}
		if nodes[to - range.start].visited {
			let mut path = vec![];
			path.reserve(nodes[to - range.start].distance + 1);
			path.push(to);
			let mut next = nodes[to - range.start].previous;
			while let Some(current) = next {
				path.push(current);
				next = nodes[current - range.start].previous;
			}
			let mut visited = vec![];
			for node in nodes {
				if node.visited && node.previous.is_some() {
					visited.push((node.id, node.previous.unwrap()));
				}
			}
			(visited, path)
		} else {
			(vec![], vec![])
		}
	}
}
